import cv2
import numpy as np
from numpy.core.numeric import identity
from numpy.lib.function_base import copy
from numpy.lib.type_check import real
from numpy.lib.utils import source
import open3d as o3d
import glob
import yaml
import math

# Main script. Calibrates projector and camera. Process images. Generates pointclouds (pcl) and meshes.

# Camera 1 is camera. Camera 2 is projector.

# Base parameters
scan_patterns_horizontal = 9        # number of picture patterns with horizontal stripes (for default projector resolution = 9)
scan_patterns_vertical = 10         # number of picture patterns with horizontal stripes (for default projector resolution = 10)
mask_treshold = 5                   # Práh osvětlení
white_black_treshold = 0            # Where to look for dependencies (Default 0)
width = 800                         # Camera resolution
height = 600
projector_width = 640               # Projector resolution
projector_height = 360
dimension = 25.0                    # Calibration chessboard size in mm (side of square) -- it affects scale of scans

visual = False                      # (Default: False) More visuals
grey_code = True                    # True -> images captured with grey code sequence (False -> binary code)
smart_registration = True           # (Default: True) True -> scans merged to two halves, then merged together, False -> added one by one (risk that last one will not fit into the scan)

scanned_object = "Box_tma"              # Folder with images of the scanned object
base_folder = "BP"                  # Folder with folders of objects
calib_folder = "Calibration_herna1" # Folder with calibration images for projector calibration
scan_version = 2                    # number - avoid name collision

camera_calib_folder = "../scans/camera_calib/"  # here save at least 10 pictures taken by camera (with calibration pattern)
projector_calib_folders = ["../"+ base_folder +"/"+ calib_folder +"/calib1/", "../"+ base_folder +"/"+ calib_folder +"/calib2/",
"../"+ base_folder +"/"+ calib_folder +"/calib3/" ,"../"+ base_folder +"/"+ calib_folder +"/calib4/" ,"../"+ base_folder +"/"+ calib_folder +"/calib5/",  
"../"+ base_folder +"/"+ calib_folder +"/calib6/" ,"../"+ base_folder +"/"+ calib_folder +"/calib7/" ,"../"+ base_folder +"/"+ calib_folder +"/calib8/"]
# "../"+ base_folder +"/"+ calib_folder +"/calib9/" ] #,"../"+ base_folder +"/"+ calib_folder +"/calib10/"


obj_folder_0 = "../"+ base_folder +"/" + scanned_object + "/" + scanned_object + "_0/"
obj_folder_1 = "../"+ base_folder +"/" + scanned_object + "/" + scanned_object + "_1/"
obj_folder_2 = "../"+ base_folder +"/" + scanned_object + "/" + scanned_object + "_2/"
obj_folder_3 = "../"+ base_folder +"/" + scanned_object + "/" + scanned_object + "_3/"
obj_folder_4 = "../"+ base_folder +"/" + scanned_object + "/" + scanned_object + "_4/"
obj_folder_5 = "../"+ base_folder +"/" + scanned_object + "/" + scanned_object + "_5/"
obj_folder_6 = "../"+ base_folder +"/" + scanned_object + "/" + scanned_object + "_6/"
obj_folder_7 = "../"+ base_folder +"/" + scanned_object + "/" + scanned_object + "_7/"

adittional_folder_1 = "../"+ base_folder +"/" + scanned_object + "/" + scanned_object + "_8/"
adittional_folder_2 = "../"+ base_folder +"/" + scanned_object + "/" + scanned_object + "_9/"
adittional_folder_3 = "../"+ base_folder +"/" + scanned_object + "/" + scanned_object + "_10/"
adittional_folder_4 = "../"+ base_folder +"/" + scanned_object + "/" + scanned_object + "_11/"
adittional_folder_5 = "../"+ base_folder +"/" + scanned_object + "/" + scanned_object + "_12/"
adittional_folder_6 = "../"+ base_folder +"/" + scanned_object + "/" + scanned_object + "_13/"

# ----------------------------------------------------------------------------------------------------------------------

def calibrate_camera():
    # Defining the dimensions of checkerboard
    CHECKERBOARD = (6,9)
    criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001)

    # Creating vector to store vectors of 3D points for each checkerboard image
    objpoints = []

    # Creating vector to store vectors of 2D points for each checkerboard image
    imgpoints = []

    objp = []

    for i in reversed(range(9)):
        for j in range(6):
            objp.append([i*dimension, j*dimension, 0]) # dimensions affect scale of scans

    objp = np.array(objp)

    # Get file directory of camera calibration images
    images = glob.glob('../scans/camera_calib/*')       # camera calibration images folder
    for file_name in images:
        img = cv2.imread(file_name)
        gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        
        # Find the chess board corners
        # If desired number of corners are found in the image then ret = true
        ret, corners = cv2.findChessboardCorners(gray, CHECKERBOARD, cv2.CALIB_CB_ADAPTIVE_THRESH + cv2.CALIB_CB_FAST_CHECK + cv2.CALIB_CB_NORMALIZE_IMAGE)
        
        # if all corners detected
        if ret == True:
            objpoints.append(objp)
            
            # Refining pixel coordinates for given 2D points.
            corners2 = cv2.cornerSubPix(gray, corners, (11,11),(-1,-1), criteria)
            imgpoints.append(corners2)

            # Draw and display the corners
            img = cv2.drawChessboardCorners(img, CHECKERBOARD, corners2, ret)

        # if visual: 
        #     cv2.imshow('img', img)
        #     cv2.waitKey(0)

    cv2.destroyAllWindows()

    h,w = img.shape[:2]
    objpoints = np.array(objpoints, dtype=np.float32)

    ret, mtx, dist, rvecs, tvecs = cv2.calibrateCamera(objpoints, imgpoints, gray.shape[::-1], None, None)
    return ret, mtx, dist, rvecs, tvecs

# ----------------------------------------------------------------------------------------------------------------------

# calibrates camera and projector
def calibrate_camera_projector(mask_treshold, camera_calib_folder, projector_calib_folders):
    ret, cam_mtx, cam_dist, cam_rvecs, cam_tvecs = calibrate_camera()
    print("Camera calibration error:")
    print(ret)
    print("")

    newcameramtx, roi = cv2.getOptimalNewCameraMatrix(cam_mtx, cam_dist, (width, height),1, (width, height))
    mapxc, mapyc = cv2.initUndistortRectifyMap(cam_mtx, cam_dist, None, newcameramtx, (width, height), 5)

    criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 40, 0.00001)
    
    # get ideal chessboard coordinates
    chessboard_planar = []
    for i in range(9):
        for j in reversed(range(6)):
            chessboard_planar.append([32 + i*72, 32 + j*72])
    chessboard_planar = np.array(chessboard_planar)
    chessboard_planar = np.flip(chessboard_planar, 0)

    projector_real_world = []
    projector_image = []
    camera_image = []

    # get coresponding points
    for folder in projector_calib_folders:
        real_world_points = []
        proj_points = []
        cam_points = []
        
        # get picture coordinates
        img = get_image(0, folder, True, True)  # black image
        gray_camera = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        gray = cv2.remap(gray_camera, mapxc, mapyc, cv2.INTER_LINEAR)
        # cv2.imshow("img", gray)
        # cv2.waitKey(0)
        ret, corners = cv2.findChessboardCorners(
            gray, (6,9), cv2.CALIB_CB_ADAPTIVE_THRESH + cv2.CALIB_CB_FAST_CHECK + cv2.CALIB_CB_NORMALIZE_IMAGE)
        corners = cv2.cornerSubPix(gray, corners, (11,11), (-1,-1), criteria)

        # find homography
        hom_mtx, _ = cv2.findHomography(corners, chessboard_planar)

        # get camera projector pixel mapping
        mask = get_mask(mask_treshold, folder, mapxc, mapyc)
        pixels = np.zeros([height, width, 3], dtype=np.int32)
        pixels = segment(pixels, mask, folder, mapxc, mapyc, scan_patterns_vertical, vertical_lines=True)
        pixels = segment(pixels, mask, folder, mapxc, mapyc, scan_patterns_horizontal, vertical_lines=False)
        if True:            # grey_code
            to_bin_parallel(pixels, mask)

        if visual:
            # show chessboard corners
            draw_img = np.copy(img)
            cv2.drawChessboardCorners(draw_img, (9,6), corners, ret)
            # show warped perspective according to homography
            img_warped = cv2.warpPerspective(draw_img, hom_mtx, (img.shape[1], img.shape[0]))
            cv2.imshow("original", draw_img)   
            cv2.imshow("homography", img_warped)
            # visualize camera projector pixel mapping
            show_pixels_x = np.copy(pixels[:,:,0])
            show_pixels_x = show_pixels_x/(2 ** (scan_patterns_vertical))*255
            show_pixels_x = show_pixels_x.astype(np.uint8)
            show_pixels_y = np.copy(pixels[:,:,1])
            show_pixels_y = show_pixels_y/(2 ** (scan_patterns_horizontal))*255
            show_pixels_y = show_pixels_y.astype(np.uint8)
            pixels_warped = cv2.warpPerspective(show_pixels_x, hom_mtx, (mask.shape[1], mask.shape[0]))
            cv2.imshow("pixels warped", pixels_warped)
            cv2.imshow("pixels_x", show_pixels_x)
            cv2.imshow("pixels_y",show_pixels_y)
            cv2.drawChessboardCorners(pixels_warped, (9,6), corners, ret)
            # close windows on keypress
            cv2.waitKey(0)
            cv2.destroyAllWindows()

        # create projector - real world pairs
        pixels_realworld = np.empty((height, width, 2))
        pixels_realworld[:,:,0] = cv2.warpPerspective(pixels[:,:,0].astype(np.uint16), hom_mtx, (width, height))
        pixels_realworld[:,:,1] = cv2.warpPerspective(pixels[:,:,1].astype(np.uint16), hom_mtx, (width, height))
        mask_warped = cv2.warpPerspective(mask, hom_mtx, (mask.shape[1], mask.shape[0]))

        # get projection of the image to the flat wall
        gray_warped = cv2.warpPerspective(gray, hom_mtx, (img.shape[1], img.shape[0]))

        ret, corners_warped = cv2.findChessboardCorners(
            gray_warped, (6,9), cv2.CALIB_CB_ADAPTIVE_THRESH + cv2.CALIB_CB_FAST_CHECK + cv2.CALIB_CB_NORMALIZE_IMAGE)
        corners_warped = cv2.cornerSubPix(gray_warped, corners_warped, (11,11), (-1,-1), criteria)

        # show warped images
        if visual:
            realworld_pixels = np.copy(pixels_realworld[:,:,0])
            realworld_pixels = realworld_pixels/(2 ** (scan_patterns_vertical))*255
            realworld_pixels = realworld_pixels.astype(np.uint8)
            cv2.drawChessboardCorners(realworld_pixels, (9,6), corners_warped, ret)
            cv2.imshow("realworld pixels", realworld_pixels)
            # cv2.imshow("gray pixel coordinates for given 2d points warped", gray_warped)
            cv2.imshow("mask warped", mask_warped)

        # get camera corners
        ret, camera_corners = cv2.findChessboardCorners(
            gray_camera, (6,9), cv2.CALIB_CB_ADAPTIVE_THRESH + cv2.CALIB_CB_FAST_CHECK + cv2.CALIB_CB_NORMALIZE_IMAGE)
        camera_corners = cv2.cornerSubPix(gray_camera, camera_corners, (11,11), (-1,-1), criteria)

        # show corners
        if visual:
            show_camera_corners = np.copy(gray_camera)
            cv2.drawChessboardCorners(show_camera_corners, (9,6), camera_corners, ret)
            cv2.imshow("camera corners",show_camera_corners)
            cv2.waitKey(0)
            cv2.destroyAllWindows()

        # create arrays of coresponding points, only points in mask
        cornerit = iter(corners_warped)
        for i in reversed(range(9)):
            for j in range(6):
                corner = next(cornerit)
                if in_mask(int(corner[0][1]), int(corner[0][0]), mask_warped):
                    real_world_points.append([i*dimension, j*dimension, 0])
                    proj_points.append(pixels_realworld[int(corner[0][1])] [int(corner[0][0])])
                    cam_points.append(camera_corners[(8 - i)*6+j][0])

        if not real_world_points:
            continue

        # change to correct format - python array -> numpy array
        projector_real_world.append(np.array(real_world_points,dtype=np.float32))
        projector_image.append(np.array(proj_points,dtype=np.float32))
        camera_image.append(np.array(cam_points,dtype=np.float32))


    projector_real_world = np.array(projector_real_world, dtype=object)
    projector_image = np.array(projector_image, dtype=object)
    camera_image = np.array(camera_image, dtype=object) 

    # calibrate projector
    ret, proj_mat, proj_dist, proj_rvec, proj_tvec = cv2.calibrateCamera(
        projector_real_world, projector_image, (projector_width, projector_height), None, None)
    print("Projector calibration error:")
    print(ret)
    print("")

    # stereo calibrate
    retval, cameraMatrix1, distCoeffs1, cameraMatrix2, distCoeffs2, R, T, E, F = cv2.stereoCalibrate(
        projector_real_world, camera_image, projector_image, cam_mtx, cam_dist, proj_mat, proj_dist, (800,600),
        flags= cv2.CALIB_USE_INTRINSIC_GUESS) 

    print("stereo calibration error: ")
    print(retval)
    print("")

    newprojmtx, roi = cv2.getOptimalNewCameraMatrix(
        cameraMatrix2, distCoeffs2, (projector_width, projector_height),1, (projector_width, projector_height))
    mapxp, mapyp = cv2.initUndistortRectifyMap(
        cameraMatrix2, distCoeffs2, None, newprojmtx, (projector_width, projector_height), 5)

    data = {'camera_matrix':    np.asarray(cameraMatrix1).tolist(),
        'camera_distortion':    np.asarray(distCoeffs1).tolist(),
        'projector_matrix':     np.asarray(cameraMatrix2).tolist(),
        'projector_distortion': np.asarray(distCoeffs2).tolist(),
        'idealCamera':          np.asarray(newcameramtx).tolist(),
        'idealProjector':       np.asarray(newprojmtx).tolist(),
        'cam1to2R':             np.asarray(R).tolist(),
        'cam1to2T':             np.asarray(T).tolist(),
        }

    # and save it to a file
    with open("calibration.yaml", "w") as f:
        yaml.dump(data, f)
    
    return cameraMatrix1, distCoeffs1, cameraMatrix2, distCoeffs2, R, T, E, F

# ----------------------------------------------------------------------------------------------------------------------

def load_coeffs():
    with open("calibration.yaml") as f:
        data = yaml.full_load(f)

    cam_mtx     = np.array(data['camera_matrix'],       dtype=np.float32)
    cam_dist    = np.array(data['camera_distortion'],   dtype=np.float32)
    proj_mtx    = np.array(data['projector_matrix'],    dtype=np.float32)
    proj_dist   = np.array(data['projector_distortion'],dtype=np.float32)
    idealCamera = np.array(data['idealCamera'],         dtype=np.float32)
    idealProj   = np.array(data['idealProjector'],      dtype=np.float32)
    cam1To2Rmat = np.array(data['cam1to2R'],            dtype=np.float32)
    cam1To2Tvec = np.array(data['cam1to2T'],            dtype=np.float32)

    return cam_mtx, cam_dist, proj_mtx, proj_dist, idealCamera, idealProj, cam1To2Rmat, cam1To2Tvec

# ----------------------------------------------------------------------------------------------------------------------

# loads image according to parameters (number of stripes/if inverse stripes, etc.)
def get_image(pattern_number, image_folder, is_vertical = True, is_inverted = False):
    if pattern_number == 0:
        if (is_inverted):
            whatever = str(2)
        else:
            whatever = str(1)
    else:
        if is_vertical:
            if is_inverted:
                whatever = str(pattern_number * 2 + 2)
            else:
                whatever = str(pattern_number * 2 + 1)
        else:
            if is_inverted:
                whatever = str(pattern_number * 2 + 22)
            else:
                whatever = str(pattern_number * 2 + 21)
    img = cv2.imread(image_folder + whatever + ".jpg")
    return img

# ----------------------------------------------------------------------------------------------------------------------

# mask to determine where is not shadow
def get_mask(treshold, object_folder, undist_map_x_cam, undist_map_y_cam):
    mask = np.zeros([height, width, 1], dtype = np.uint8)
                
    # projector black
    img_black = get_image(0, object_folder, is_inverted = True)
    img_black = cv2.remap(img_black, undist_map_x_cam, undist_map_y_cam, cv2.INTER_LINEAR)
    
    # projector white
    img_white = get_image(0, object_folder, is_inverted = False)
    img_white = cv2.remap(img_white, undist_map_x_cam, undist_map_y_cam, cv2.INTER_LINEAR)

    img_black = cv2.cvtColor(img_black, cv2.COLOR_BGR2GRAY)     # to grayscale
    img_white = cv2.cvtColor(img_white, cv2.COLOR_BGR2GRAY)     # to grayscale

    # img_black - img_white
    for i in range(height):
        for j in range(width):
            if (int(img_white[i][j]) - int(img_black[i][j]) > treshold):
                mask[i][j][0] = 255
    
    if visual:
        cv2.imshow("black mask", img_black)
        cv2.imshow("white mask", img_white)
        cv2.imshow("mask", mask)
        cv2.waitKey(0)
        cv2.destroyAllWindows()
    
    return mask

# ----------------------------------------------------------------------------------------------------------------------

# returns array in shape ()
def segment(pixels, mask, object_folder, undist_map_x_cam, undist_map_y_cam, scan_patterns, vertical_lines = True):
    for l in range(scan_patterns):
        # get image
        img = get_image(l + 1, object_folder, is_vertical = vertical_lines, is_inverted = False)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        img = cv2.remap(img, undist_map_x_cam, undist_map_y_cam, cv2.INTER_LINEAR)
        
        # get inverted image
        img_inv = get_image(l + 1, object_folder, is_vertical = vertical_lines, is_inverted = True)
        img_inv = cv2.cvtColor(img_inv, cv2.COLOR_BGR2GRAY)
        img_inv = cv2.remap(img_inv, undist_map_x_cam, undist_map_y_cam, cv2.INTER_LINEAR)

        # do one layer
        difference = img.astype(np.int16) - img_inv.astype(np.int16)
        if (vertical_lines):
            pixels[:, :, 0] |= np.piecewise(difference, [difference > white_black_treshold], [1 << (scan_patterns - l - 1), 0])
            if l == scan_patterns - 2:
                pixels[:, :, 2] = img
        else:
            pixels[:, :, 1] |= np.piecewise(difference, [difference > white_black_treshold], [1 << (scan_patterns - l - 1), 0])

    return pixels

# ----------------------------------------------------------------------------------------------------------------------

# determines if is the point in mask (if projector emmits light to this place of picture)
def in_mask(i, j, mask):
    if mask[i][j] == 255 :
        return True
    else :
        return False

# ----------------------------------------------------------------------------------------------------------------------

# grey code to bin code computating
def to_bin(pixels, mask):
    for i in range(height):
        for j in range(width):
            if in_mask(i, j, mask):
                pixels[i][j][0] = gray_to_bin(pixels[i][j][0], scan_patterns_vertical)
                pixels[i][j][1] = gray_to_bin(pixels[i][j][1], scan_patterns_horizontal)

# ----------------------------------------------------------------------------------------------------------------------

# fast version of grey code to bin code computating
def to_bin_parallel(pixels, mask):
    pixels[:,:,0] = gray_to_bin_parallel(pixels[:,:,0], scan_patterns_vertical)
    pixels[:,:,1] = gray_to_bin_parallel(pixels[:,:,1], scan_patterns_horizontal)

# ----------------------------------------------------------------------------------------------------------------------

# grey code to bin code computating           
def gray_to_bin(number, bits):
    value = number & (1 << (bits-1))
    for b in range(1, bits):
        value |= ((value & (1 << (bits - b))) >> 1) ^ (number & (1 << (bits - b - 1)))
    return value

# ----------------------------------------------------------------------------------------------------------------------

# fast version of grey code to bin code computating
def gray_to_bin_parallel(number, bits):
    value = np.empty([height, width, 2], dtype=np.int32)
    value = number & (1 << (bits-1))
    for b in range(1, bits):
        value |= ((value & (1 << (bits - b))) >> 1) ^ (number & (1 << (bits - b - 1)))
        
    return value

# ----------------------------------------------------------------------------------------------------------------------

# find aproximate intersection of two rays
def ray_intersect(p1, v1, p2, v2):
    v1_sqr_sum = 0.0
    v2_sqr_sum = 0.0
    v1_v2_sum  = 0.0
    q12_v1_sum = 0.0
    q12_v2_sum = 0.0
    for i in range(3):
        q12 = p1[i][0] - p2[i][0]
        v1_sqr_sum += v1[i] * v1[i]
        v2_sqr_sum += v2[i] * v2[i]
        v1_v2_sum  += v1[i] * v2[i]
        q12_v1_sum += q12 * v1[i]
        q12_v2_sum += q12 * v2[i]

    denom = v1_sqr_sum * v2_sqr_sum - v1_v2_sum * v1_v2_sum
    s = (v1_v2_sum/denom)*q12_v2_sum - (v2_sqr_sum/denom)*q12_v1_sum
    t = -(v1_v2_sum/denom)*q12_v1_sum + (v1_sqr_sum/denom)*q12_v2_sum

    p = np.empty([3], np.float)

    for i in range(3):
        p[i] = ((p1[i][0] + s*v1[i]) + (p2[i][0] + t * v2[i]))/2

    d_sqr = 0
    for i in range(3):
        d_sqr += ((p1[i][0] + s*v1[i]) - (p2[i][0] + t * v2[i])) ** 2
    d = math.sqrt(d_sqr)

    # delete wrong points
    if d > 10:
        p = ([0, 0, 0])
        d = 0

    return p, d

# ----------------------------------------------------------------------------------------------------------------------

# adds normals to pcls, normals are used for mesh generation
def add_normals(pcl):
    pcl.estimate_normals(search_param=o3d.geometry.KDTreeSearchParamHybrid(radius=0.1, max_nn=30))
    normals = np.asarray(pcl.normals)
    pcl.normals = o3d.utility.Vector3dVector(-normals)
    return pcl

# ----------------------------------------------------------------------------------------------------------------------

# visualizes registration
def draw_registration_result(source, target):
    source_temp = source
    target_temp = target
    # source_temp.paint_uniform_color([1, 0.706, 0])
    # target_temp.paint_uniform_color([0, 0.651, 0.929])
    o3d.visualization.draw_geometries([source_temp, target_temp])

# ----------------------------------------------------------------------------------------------------------------------

# pick points to be used in registration
def pick_points(pcd):
    print("")
    print("1) Please pick at least three correspondences using [shift + left click]")
    print("   Press [shift + right click] to undo point picking")
    print("2) After picking points, press 'Q' to close the window")
    vis = o3d.visualization.VisualizerWithEditing()
    vis.create_window()
    vis.add_geometry(pcd)
    vis.run()  # user picks points
    vis.destroy_window()
    print("")

    return vis.get_picked_points()

# ----------------------------------------------------------------------------------------------------------------------

# manual registration to 2 halves, which are than joined together (minimalizes inaccuracies)
def manual_registration_smart(pcl_1, pcl_2):
    target = pcl_1
    pcl_combined = pcl_2

    # pick points from two point clouds and builds correspondences
    picked_id_target = pick_points(target)
    picked_id_source = pick_points(pcl_combined)
    assert (len(picked_id_source) >= 3 and len(picked_id_target) >= 3)
    assert (len(picked_id_source) == len(picked_id_target))
    corr = np.zeros((len(picked_id_source), 2))
    corr[:, 0] = picked_id_source
    corr[:, 1] = picked_id_target

    # estimate rough transformation using correspondences
    print("Compute a rough transform using the correspondences given by user")
    p2p = o3d.pipelines.registration.TransformationEstimationPointToPoint()
    trans_init = p2p.compute_transformation(pcl_combined, target, o3d.utility.Vector2iVector(corr))

    # point-to-point ICP for refinement
    print("Perform point-to-point ICP refinement")
    threshold = 0.03  # 3cm distance threshold
    reg_p2p = o3d.pipelines.registration.registration_icp(
        pcl_combined, target, threshold, trans_init,
        o3d.pipelines.registration.TransformationEstimationPointToPoint())
    pcl_combined.transform(reg_p2p.transformation)

    draw_registration_result(pcl_combined, target)
    print("")

    return pcl_combined, target

# ----------------------------------------------------------------------------------------------------------------------

# manual registration
def manual_registration(pcl, counter):
    pcl_combined = load_cropped(counter + 1)
    target = pcl

    # pick points from two point clouds and builds correspondences
    picked_id_target = pick_points(target)
    picked_id_source = pick_points(pcl_combined)
    assert (len(picked_id_source) >= 3 and len(picked_id_target) >= 3)
    assert (len(picked_id_source) == len(picked_id_target))
    corr = np.zeros((len(picked_id_source), 2))
    corr[:, 0] = picked_id_source
    corr[:, 1] = picked_id_target

    # estimate rough transformation using correspondences
    print("Compute a rough transform using the correspondences given by user")
    p2p = o3d.pipelines.registration.TransformationEstimationPointToPoint()
    trans_init = p2p.compute_transformation(pcl_combined, target, o3d.utility.Vector2iVector(corr))

    # point-to-point ICP for refinement
    print("Perform point-to-point ICP refinement")
    threshold = 0.03  # 3cm distance threshold
    reg_p2p = o3d.pipelines.registration.registration_icp(
        pcl_combined, target, threshold, trans_init,
        o3d.pipelines.registration.TransformationEstimationPointToPoint())
    pcl_combined.transform(reg_p2p.transformation)

    draw_registration_result(pcl_combined, target)
    print("")

    return pcl_combined

# ----------------------------------------------------------------------------------------------------------------------

# registration
def execute_global_registration(source_down, target_down, source_fpfh, target_fpfh, voxel_size):
    distance_threshold = voxel_size * 1.5
    print(":: RANSAC registration on downsampled point clouds.")
    print("   Since the downsampling voxel size is %.3f," % voxel_size)
    print("   we use a liberal distance threshold %.3f." % distance_threshold)
    result = o3d.pipelines.registration.registration_ransac_based_on_feature_matching(
        source_down, target_down, source_fpfh, target_fpfh, True, distance_threshold,
        o3d.pipelines.registration.TransformationEstimationPointToPoint(False),
        3, [
            o3d.pipelines.registration.CorrespondenceCheckerBasedOnEdgeLength(
                0.9),
            o3d.pipelines.registration.CorrespondenceCheckerBasedOnDistance(
                distance_threshold)
        ], o3d.pipelines.registration.RANSACConvergenceCriteria(100000, 0.999))
    return result

# ----------------------------------------------------------------------------------------------------------------------

# loads manualy cropped images from folder
def load_cropped(i):
    cropped_1 = o3d.io.read_point_cloud("cropped/" + scanned_object + "_" + str(i+1) + ".ply")
    cropped_1 = cropped_1.voxel_down_sample(voxel_size=0.7)
    cropped_1 = add_normals(cropped_1)

    return cropped_1

# ----------------------------------------------------------------------------------------------------------------------

# transforms pcl according to transformation got from registration
def transform_pcl(pcl, transformation):
    for i in range(len(transformation)):
        for j in range(len(transformation)-i):
            pcl[j+i].transform(transformation[i])
    
    return pcl

# ----------------------------------------------------------------------------------------------------------------------

# registrates pcls in two halves and then joins them into one - lowering inacurracies
def combine_pcl_smart(pcl):
    pcl_combined_1 = pcl[0]
    pcl_combined_2 = pcl[4]
    for i in range(int(math.ceil((len(pcl)/2)-1))):
        pcl_combined_1 += pcl[i+1]

    for i in range(int(math.ceil((len(pcl)/2)-1))):
        pcl_combined_2 += pcl[i+5]

    o3d.visualization.draw_geometries([pcl_combined_1, pcl_combined_2])

    return pcl_combined_1, pcl_combined_2

# ----------------------------------------------------------------------------------------------------------------------

# registrates pcls and than merges 2 pcls into 1 pcl
def combine_pcl(pcl):
    pcl_combined = pcl[0]
    for i in range(len(pcl)-1):
        pcl_combined += pcl[i+1]

    o3d.visualization.draw_geometries([pcl_combined])

    return pcl_combined

# ----------------------------------------------------------------------------------------------------------------------

# function can be called by user to get more accurate registration (can also damage registration)
def ICP_again(pcl_combined, target):
    threshold = 0.2
    trans_init = np.identity(4)
    trans_init = o3d.pipelines.registration.registration_icp(
            pcl_combined, target, threshold, trans_init,
            o3d.pipelines.registration.TransformationEstimationPointToPoint())
    for i in range(5):
        threshold = 0.1
        trans_init = o3d.pipelines.registration.registration_icp(
            pcl_combined, target, threshold, trans_init.transformation,
            o3d.pipelines.registration.TransformationEstimationPointToPoint())
        pcl_combined.transform(trans_init.transformation)
        threshold = threshold/1.5
    draw_registration_result(pcl_combined, target)

    return pcl_combined

# ---------------------------------------------------------------------------------------------------------------------

# create pcl from pictures
def Images_to_pcl(obj_folder, turn_counter):
    # get undistort map
    mapxc, mapyc = cv2.initUndistortRectifyMap(camera_mtx, camera_dist, None, newcameramtx, (width, height), 5)

    mask = get_mask(mask_treshold, obj_folder, mapxc, mapyc)
    pixels = np.zeros([height, width, 3], dtype=np.int32)
    pixels = segment(pixels, mask, obj_folder, mapxc, mapyc, scan_patterns_vertical, vertical_lines=True)
    pixels = segment(pixels, mask, obj_folder, mapxc, mapyc, scan_patterns_horizontal, vertical_lines=False)
    if grey_code:
        to_bin_parallel(pixels, mask)
    
    if visual:
        show_pixels_x = np.copy(pixels[:,:,0])
        show_pixels_x = show_pixels_x/(2 ** (scan_patterns_vertical))*255
        show_pixels_x = show_pixels_x.astype(np.uint8)
        show_pixels_y = np.copy(pixels[:,:,1])
        show_pixels_y = show_pixels_y/(2 ** (scan_patterns_horizontal))*255
        show_pixels_y = show_pixels_y.astype(np.uint8)
        cv2.imshow("pixels_x", show_pixels_x)
        cv2.imshow("pixels_y",show_pixels_y)
        cv2.waitKey(0)
        cv2.destroyAllWindows()

    # create np array representing pointcloud
    np_pointcloud = np.zeros([width*height,6], np.float32) # np.empty([0,6], np.float32)
    visualization = []
    line_set = o3d.geometry.LineSet(
        points = o3d.utility.Vector3dVector([[0.0,0.0,0.0], [0.0, 0.0, 100.0], [0.0,100.0,0.0],[100.0,0.0,0.0]]),
        lines  = o3d.utility.Vector2iVector([[0,1], [0,2], [0,3]])
    )
    visualization.append(line_set)
    pcl_points = 0

    # transfer rotation vectors to rotation matrixes and get inverse matrixes
    # homogenous coordinates
    coord_transform_matrix = np.c_[cam1To2Rmat, cam1To2Tvec]
    coord_transform_matrix = np.r_[coord_transform_matrix, [[0,0,0,1]]]

    camera_mtx_hom = np.c_[camera_mtx, [[0],[0],[0]]]
    camera_mtx_hom = np.r_[camera_mtx_hom, [[0,0,0,1]]]
    camera_mtx_hom_inv = np.linalg.inv(camera_mtx_hom)

    newcamera_mtx_hom = np.c_[newcameramtx, [[0],[0],[0]]]
    newcamera_mtx_hom = np.r_[newcamera_mtx_hom, [[0,0,0,1]]]
    newcamera_mtx_hom_inv = np.linalg.inv(newcamera_mtx_hom)

    newprojmtx_hom = np.c_[newprojmtx, [[0],[0],[0]]]
    newprojmtx_hom = np.r_[newprojmtx_hom, [[0,0,0,1]]]
    newprojmtx_hom_inv = np.linalg.inv(newprojmtx_hom)

    projmtx_hom = np.c_[proj_mtx, [[0],[0],[0]]]
    projmtx_hom = np.r_[projmtx_hom, [[0,0,0,1]]]
    projmtx_hom_inv = np.linalg.inv(projmtx_hom)

    # get image for color
    color_img = get_image(0, obj_folder)
    color_img = cv2.remap(color_img, mapxc, mapyc, cv2.INTER_LINEAR)

    ray_counter = 0
    distance_arr = np.array([])

    # get 3d points and add them to pointcloud
    for i in range(height):
        for j in range(width):
            if in_mask(i, j, mask):
                # camera origin and vectors
                proj_x = pixels[i][j][0]
                proj_y = pixels[i][j][1]
                # camera origin and vector
                camera_point = np.matmul(coord_transform_matrix ,np.array([[0.0],[0.0],[0.0],[1.0]]))   # np.array([[0.0],[0.0],[0.0]]) - cam1To2Tvec
                cam_coords_vec_point = np.matmul(newcamera_mtx_hom_inv, np.array([[j], [i], [1], [1]])) # np.matmul(newcameramtx_inv, np.array([[j], [i], [1]]))
                camera_vec = (np.matmul(coord_transform_matrix, cam_coords_vec_point) - camera_point).T[0]
                # projector origin and vector
                proj_point = np.array([[0.0], [0.0], [0.0], [1.0]])
                proj_coords_vec_point = np.matmul(newprojmtx_hom_inv, np.array([[proj_x], [proj_y], [1], [1]]))
                proj_vec = proj_coords_vec_point.T[0]

                # rays intersection
                intersection, distance = ray_intersect(proj_point, proj_vec, camera_point, camera_vec)
                distance_arr = np.append(distance_arr, distance)

                if visual and ray_counter%100 == 0:
                    # create lines
                    line_set = o3d.geometry.LineSet(
                        points = o3d.utility.Vector3dVector([
                            camera_point.T[0][:3],
                            (camera_point.T[0] + camera_vec*1000)[:3],
                            proj_point.T[0][:3],
                            (proj_point.T[0] + proj_vec*1000)[:3]
                        ]),
                        lines = o3d.utility.Vector2iVector([[0,1], [2,3]])
                    )
                    intersections = o3d.geometry.PointCloud()
                    intersections.points = o3d.utility.Vector3dVector([intersection])
                    visualization.append(line_set)
                    visualization.append(intersections)
                    # if ray_counter % 1000 == 0:
                    # o3d.visualization.draw_geometries(visualization)
                
                # add to pointcloud
                data = np.append(intersection, np.flip(color_img[i][j]))
                np_pointcloud[i*height + j] = data # np_pointcloud = np.append(np_pointcloud, data)

                ray_counter = ray_counter + 1

    # filtering 0 points
    np_pointcloud = np_pointcloud[~np.any(np_pointcloud == 0, axis=1)]

    if visual:
        o3d.visualization.draw_geometries(visualization)

    # init pointcloud with np array
    pcl = o3d.geometry.PointCloud()
    pcl.points = o3d.utility.Vector3dVector(np_pointcloud[:,:3])
    pcl.colors = o3d.utility.Vector3dVector(np_pointcloud[:,3:6]/255)

    # filter with open3D
    pcl, ind = pcl.remove_statistical_outlier(nb_neighbors=30, std_ratio=0.3)
    pcl, ind = pcl.remove_radius_outlier(nb_points=8, radius=3)

    print('Use cursor to rotate // "k" key to lock wiev for edit // (ctrl) + mouse to choose good points // "c" key to crop and save as:')
    print('folder: /cropped')
    print('Save as: ' + scanned_object + '_' + str(turn_counter+1) + '')
    o3d.visualization.draw_geometries_with_editing([pcl])

# ----------------------------------------------------------------------------------------------------------------------
# Start of the program
# ----------------------------------------------------------------------------------------------------------------------

# load folder list with scan pictures
obj_folder_list = [
    obj_folder_0,
    obj_folder_1,
    obj_folder_2,
    obj_folder_3,
    obj_folder_4,
    obj_folder_5,
    obj_folder_6,
    obj_folder_7,
    ]

turn_counter = 0

while True:
    state = input("choose action (type one of following and confirm by ENTER):\ncalibrate, new_pcl, combine_pcl, adittional_scans, pcl_to_mesh, all, end, treshold, visualization\n")

    if state == "end":
        break

    if state == "treshold":
        mask_treshold = int(input("treshold="))

    # get camera matrixes and correction maps
    if state == "calibrate":
        calibrate_camera_projector(mask_treshold, camera_calib_folder, projector_calib_folders)

    # load parameters
    if state == "new_pcl" or state == "all":
        camera_mtx, camera_dist, proj_mtx, proj_dist, newcameramtx, newprojmtx, cam1To2Rmat, cam1To2Tvec = load_coeffs()

        # create pointcloud from pictures
        for obj_folder in obj_folder_list:
            Images_to_pcl(obj_folder, turn_counter)
            turn_counter += 1
            
    # ----------------------------------------------------------------------------------------------------------------------
    # points to mesh - load cropped, manual registration, merging pcls, create mesh

    if state == "combine_pcl" or state == "all":
        if smart_registration:
            i = 0
            pcl = {}
            pcl[0] = load_cropped(0)
            pcl[4] = load_cropped(4)

            while i < (len(obj_folder_list)-1):
                try:
                    pcl[i+1] = manual_registration(pcl[i], i)
                except:
                    input("wrong number of points, enter and retry:\n")
                    continue
                
                while True:
                    user = int(input("0:retry, 1:next, 2:ICP (accuracy)\n"))
                    if user == 0:
                        i = i
                        break
                    elif user == 1:
                        i += 1
                        if smart_registration & (i==((len(obj_folder_list)/2)-1)):
                            i += 1
                        break
                    elif user == 2:
                        pcl[i+1] = ICP_again(pcl[i+1], pcl[i])
                    
        if smart_registration:
            pcl_combined_1, pcl_combined_2 = combine_pcl_smart(pcl)
            
            i = 0
            while i < 1:
                try:
                    pcl_combined, target = manual_registration_smart(pcl_combined_1, pcl_combined_2)
                except:
                    input("wrong number of points, enter and retry:\n")
                    continue

                while True:
                        user = int(input("0:retry, 1:next, 2:ICP (accuracy)\n"))
                        if user == 0:
                            i = i
                            break
                        elif user == 1:
                            i += 1
                            pcl_combined += target
                            break
                        elif user == 2:
                            pcl[i+1] = ICP_again(pcl_combined, target)

        # not in default mode
        else:
            pcl_combined = combine_pcl(pcl)

        # mesh downsample
        pcl_combined = pcl_combined.voxel_down_sample(voxel_size=0.7)

        # save pcl_combined
        o3d.io.write_point_cloud("../point_clouds/plc_combined_" + scanned_object + "_" + str(scan_version) + ".ply", pcl_combined)

    # adittional pcl registration
    if state == "adittional_scans" or state == "all":
        camera_mtx, camera_dist, proj_mtx, proj_dist, newcameramtx, newprojmtx, cam1To2Rmat, cam1To2Tvec = load_coeffs()
        pcl_combined_1 = o3d.io.read_point_cloud("../point_clouds/plc_combined_" + scanned_object + "_" + str(scan_version) + ".ply")
        count = 0
        user = int(input("0: continue to mesh\n1: adittional scan\n"))
        while user == 1:
            if count == 0:
                adittional_folder = adittional_folder_1
            elif count == 1:
                adittional_folder = adittional_folder_2
            elif count == 2:
                adittional_folder = adittional_folder_3
            else:
                adittional_folder = adittional_folder_4

            Images_to_pcl(adittional_folder, 30+count)
            pcl_add = load_cropped((30+count))

            i = 0
            while i < 1:
                try:
                    pcl_combined, target = manual_registration_smart(pcl_add, pcl_combined_1)
                except:
                    input("wrong number of points, enter and retry:\n")
                    continue

                while True:
                        user = int(input("0:retry, 1:next, 2:ICP (accuracy)\n"))
                        if user == 0:
                            i = i
                            break
                        elif user == 1:
                            i += 1
                            pcl_combined += target
                            count += 1
                            break
                        elif user == 2:
                            pcl = ICP_again(pcl_combined, target)

            user = int(input("0: continue to mesh\n1: adittional scan\n"))

        # save pcl_combined
        o3d.io.write_point_cloud("../point_clouds/plc_combined_final_" + scanned_object + "_" + str(scan_version) + ".ply", pcl_combined)

    # generate mesh
    if state == "pcl_to_mesh" or state == "all":
        pcl_combined = o3d.io.read_point_cloud("../point_clouds/plc_combined_final_" + scanned_object + "_" + str(scan_version) + ".ply")
        pcl_combined = pcl_combined.voxel_down_sample(voxel_size=0.7)

        alpha = 3
        print(f"alpha={alpha:.3f}")
        mesh = o3d.geometry.TriangleMesh.create_from_point_cloud_alpha_shape(pcl_combined, alpha)
        mesh.compute_vertex_normals()
        print('mesh generated using alfa shapes method')
        o3d.visualization.draw_geometries([mesh], mesh_show_back_face=True)

        # Mesh filtering
        mesh = mesh.filter_smooth_simple(number_of_iterations=2)
        mesh.compute_vertex_normals()
        print('filtered mesh generated by alfa shapes')
        o3d.visualization.draw_geometries([mesh])

        # ball pivoting
        radii = [0.05, 0.1, 0.2, 0.4, 0.8, 1.6, 3.2]
        rec_mesh = o3d.geometry.TriangleMesh.create_from_point_cloud_ball_pivoting(pcl_combined, o3d.utility.DoubleVector(radii))
        print('mesh generated using ball pivoting method')
        o3d.visualization.draw_geometries([rec_mesh])

        # Mesh filtering
        rec_mesh = rec_mesh.filter_smooth_simple(number_of_iterations=2)
        rec_mesh.compute_vertex_normals()
        print('filtered mesh generated by ball pivoting')
        o3d.visualization.draw_geometries([rec_mesh])

        # Mesh save
        o3d.io.write_triangle_mesh("../point_clouds/mesh_ball_" + scanned_object + "_" + str(scan_version) + ".ply", rec_mesh)
        # o3d.io.write_triangle_mesh("../point_clouds/mesh_ball_" + scanned_object + "_" + str(scan_version) + ".stl", rec_mesh) # another file option

        # poisson surface reconstruction
        with o3d.utility.VerbosityContextManager(o3d.utility.VerbosityLevel.Debug) as cm:
            mesh, densities = o3d.geometry.TriangleMesh.create_from_point_cloud_poisson(pcl_combined, depth=9)
        print(mesh)
        print('mesh generated using Poisson surface reconstruction method')
        o3d.visualization.draw_geometries([mesh])
        o3d.io.write_triangle_mesh("../BP_scans/mesh_poisson_" + scanned_object + "_" + str(scan_version) + ".ply", mesh)
        # o3d.io.write_triangle_mesh("../point_clouds/mesh_poisson_" + scanned_object + "_" + str(scan_version) + ".obj", mesh) # another file option

    if state == "visualization":
        pcl_combined = o3d.io.read_point_cloud("../BP_scans/mesh_poisson_" + scanned_object + "_" + str(scan_version) + ".ply")
        print(pcl_combined)
        o3d.visualization.draw_geometries([pcl_combined])